from .chat import get_chat_model, get_chat_model_full
from .decorator import agent
from .thread import get_default_thread

__all__ = ["agent", "get_chat_model", "get_chat_model_full", "get_default_thread"]



"""
This module provides functionality for executing HTTP requests against Beamlit resources.
"""
import urllib.parse
from typing import Any

import requests

from beamlit.client import AuthenticatedClient
from beamlit.common import HTTPError, get_settings


class RunClient:
    """Provides functionality for executing HTTP requests against Beamlit resources.

    This module contains the RunClient class which handles authenticated HTTP requests to Beamlit
    resources. It allows users to interact with different resource types (like functions or services), supporting various HTTP methods and request parameters.

    Example:
        ```python
        client = new_client()
        run_client = RunClient(client)
        response = run_client.run(
            resource_type="function",
            resource_name="my-function",
            method="POST",
            json={"key": "value"}
        )
        ```

    Args:
        client (AuthenticatedClient): An authenticated client instance for making HTTP requests.
    """

    def __init__(self, client: AuthenticatedClient):
        self.client = client

    def run(
        self,
        resource_type: str,
        resource_name: str,
        method: str,
        path: str = "",
        headers: dict[str, str] | None = None,
        json: dict[str, Any] | None = None,
        data: str | None = None,
        params: dict[str, str] | None = None,
        cloud: bool = False,
        service_name: str | None = None,
    ) -> requests.Response:
        """Execute an HTTP request against a Beamlit resource.

        Args:
            resource_type (str): The type of resource to interact with (e.g., 'function', 'service').
            resource_name (str): The name of the specific resource.
            method (str): The HTTP method to use (e.g., 'GET', 'POST', 'PUT', 'DELETE').
            path (str, optional): Additional path segments to append to the resource URL. Defaults to "".
            headers (dict[str, str] | None, optional): HTTP headers to include in the request. Defaults to None.
            json (dict[str, Any] | None, optional): JSON payload to send with the request. Defaults to None.
            data (str | None, optional): Raw data to send with the request. Defaults to None.
            params (dict[str, str] | None, optional): Query parameters to include in the URL. Defaults to None.
            cloud (bool, optional): Whether to use the cloud endpoint. Defaults to False.
            service_name (str | None, optional): The name of the service to use. Defaults to None.

        Returns:
            requests.Response: The HTTP response from the server.

        Raises:
            HTTPError: If the server responds with a status code >= 400.
        """
        settings = get_settings()
        headers = headers or {}
        params = params or {}

        if cloud and path and service_name:
            url = f"https://{service_name}.{settings.run_internal_hostname}/{path}"

        if cloud and not path and service_name:
            url = f"https://{service_name}.{settings.run_internal_hostname}"

        if not cloud and path:
            url = urllib.parse.urljoin(settings.run_url, f"{settings.workspace}/{resource_type}s/{resource_name}/{path}")

        if not cloud and not path:
            url = urllib.parse.urljoin(settings.run_url, f"{settings.workspace}/{resource_type}s/{resource_name}")

        kwargs = {
            "headers": headers,
            "params": {**params},
        }
        if data:
            kwargs["data"] = data
        if json:
            kwargs["json"] = json

        response = self.client.request(method, url, **kwargs)
        if response.status_code >= 400 and not cloud:
            raise HTTPError(response.status_code, response.text)
        if response.status_code >= 400 and cloud: # Redirect to the public endpoint if the resource is in the cloud and the request fails
            if path:
                url = urllib.parse.urljoin(settings.run_url, f"{settings.workspace}/{resource_type}s/{resource_name}/{path}")
            else:
                url = urllib.parse.urljoin(settings.run_url, f"{settings.workspace}/{resource_type}s/{resource_name}")
            response = self.client.request(method, url, **kwargs)
            if response.status_code >= 400:
                raise HTTPError(response.status_code, response.text)
            return response
        return response



from beamlit.functions import function


@function(
    function={
        "spec": {
            "policies": ["only-us"],
        }
    }
)
def helloworld(query: str):
    """A function for saying hello to the world."""
    return "Hello from Beamlit!"



"""
This module provides utility functions for string manipulation, including slugification.
The `slugify` function transforms a given string into a URL-friendly slug by replacing spaces and underscores with hyphens and converting the string to lowercase.
"""

def slugify(name: str) -> str:
    """
    Converts a given string into a URL-friendly slug.

    This function transforms the input string by converting it to lowercase and replacing spaces and underscores with hyphens.

    Parameters:
        name (str): The string to slugify.

    Returns:
        str: The slugified version of the input string.
    """
    return name.lower().replace(" ", "-").replace("_", "-")


"""
This module provides functionalities to generate and manage Beamlit deployment configurations.
It includes functions to set default deployment values, create deployment configurations from resources,
format deployments, and clean up auto-generated deployments.
"""

import ast
import json
import os
import shutil
import sys
from logging import getLogger
from pathlib import Path
from typing import Literal

import yaml
from beamlit.api.agents import get_agent
from beamlit.authentication import new_client
from beamlit.client import AuthenticatedClient
from beamlit.common import slugify
from beamlit.common.settings import Settings, get_settings, init
from beamlit.models import (Agent, AgentSpec, Flavor, Function, FunctionSpec,
                            Metadata, MetadataLabels)

from .format import arg_to_dict
from .parser import Resource, get_description, get_parameters, get_resources

sys.path.insert(0, os.getcwd())
sys.path.insert(0, os.path.join(os.getcwd(), "src"))

def set_default_values(resource: Resource, deployment: Agent | Function):
    """
    Sets default values for a deployment based on the resource and deployment type.

    Parameters:
        resource (Resource): The resource information.
        deployment (Agent | Function): The deployment instance to set defaults for.

    Returns:
        Agent | Function: The updated deployment with default values set.
    """
    settings = get_settings()
    deployment.metadata.workspace = settings.workspace
    if not deployment.metadata.name:
        deployment.metadata.name = slugify(resource.name)
    if not deployment.metadata.display_name:
        deployment.metadata.display_name = deployment.metadata.name
    if not deployment.spec.description:
        deployment.spec.description = get_description(None, resource)
    if isinstance(deployment, Agent):
        deployment.spec.functions = []
        for arg in resource.decorator.keywords:
            if arg.arg == "remote_functions":
                if isinstance(arg.value, ast.List):
                    for value in arg.value.elts:
                        if isinstance(value, ast.Constant):
                            deployment.spec.functions.append(slugify(value.value))
    return deployment

def get_beamlit_deployment_from_resource(
    settings: Settings,
    resource: Resource,
) -> Agent | Function:
    """
    Creates a deployment configuration from a given resource.

    Args:
        resource (Resource): The resource to create a deployment for.

    Returns:
        Agent | Function: The deployment configuration.
    """
    for arg in resource.decorator.keywords:
        if arg.arg == "agent":
            if isinstance(arg.value, ast.Dict):
                value = arg_to_dict(arg.value)
                metadata = Metadata(**value.get("metadata", {}))
                spec = AgentSpec(**value.get("spec", {}))
                agent = Agent(metadata=metadata, spec=spec)
                if not agent.spec.prompt:
                    agent.spec.prompt = get_description(None, resource)
                return set_default_values(resource, agent)
        if arg.arg == "function":
            if isinstance(arg.value, ast.Dict):
                value = arg_to_dict(arg.value)
                metadata = Metadata(**value.get("metadata", {}))
                spec = FunctionSpec(**value.get("spec", {}))
                func = Function(metadata=metadata, spec=spec)
                if not func.spec.parameters:
                    func.spec.parameters = get_parameters(resource)
                return set_default_values(resource, func)
    if resource.type == "agent":
        agent = Agent(metadata=Metadata(), spec=AgentSpec())
        return set_default_values(resource, agent)
    if resource.type == "function":
        func = Function(metadata=Metadata(), spec=FunctionSpec())
        func.spec.parameters = get_parameters(resource)
        return set_default_values(resource, func)
    return None

def get_flavors(flavors: list[Flavor]) -> str:
    """
    Converts a list of Flavor objects to a JSON string.

    Args:
        flavors (list[Flavor]): List of Flavor objects.

    Returns:
        str: JSON string representation of flavors.
    """
    if not flavors:
        return "[]"
    return json.dumps([flavor.to_dict() for flavor in flavors])

def get_agent_yaml(
    agent: Agent, functions: list[tuple[Resource, Function]], settings: Settings, client: AuthenticatedClient
) -> str:
    """
    Generates YAML configuration for an agent deployment.

    Args:
        agent (Agent): Agent deployment configuration
        functions (list[tuple[Resource, FunctionDeployment]]): List of associated functions
        settings (Settings): Application settings

    Returns:
        str: YAML configuration string
    """
    try:
        agent_response = get_agent.sync(agent.metadata.name, client=client)
        agent.spec.repository = agent_response.spec.repository
    except Exception:
        pass
    agent.spec.functions = agent.spec.functions or []
    agent.spec.functions = agent.spec.functions + [slugify(function.metadata.name) for (_, function) in functions]
    agent.metadata.labels = agent.metadata.labels and MetadataLabels.from_dict(agent.metadata.labels) or MetadataLabels()
    agent.metadata.labels["x-beamlit-auto-generated"] = "true"
    agent_yaml = yaml.dump(agent.to_dict())
    template = f"""
apiVersion: beamlit.com/v1alpha1
kind: Agent
{agent_yaml}
"""
    return template


def get_function_yaml(function: Function, settings: Settings, client: AuthenticatedClient) -> str:
    """
    Generates YAML configuration for a function deployment.

    Args:
        function (FunctionDeployment): Function deployment configuration
        settings (Settings): Application settings

    Returns:
        str: YAML configuration string
    """
    function.metadata.labels = function.metadata.labels and MetadataLabels.from_dict(function.metadata.labels) or MetadataLabels()
    function.metadata.labels["x-beamlit-auto-generated"] = "true"
    function_yaml = yaml.dump(function.to_dict())
    return f"""
apiVersion: beamlit.com/v1alpha1
kind: Function
{function_yaml}
"""


def dockerfile(
    type: Literal["agent", "function"],
    resource: Resource,
    deployment: Agent | Function,
) -> str:
    """
    Generates Dockerfile content for agent or function deployment.

    Args:
        type (Literal["agent", "function"]): Type of deployment
        resource (Resource): Resource to be deployed
        deployment (Agent | Function): Resource configuration

    Returns:
        str: Dockerfile content
    """
    settings = get_settings()
    if type == "agent":
        module = f"{resource.module.__file__.split('/')[-1].replace('.py', '')}.{resource.module.__name__}"
    else:
        module = f"functions.{resource.module.__file__.split('/')[-1].replace('.py', '')}.{resource.module.__name__}"
    cmd = ["bl", "serve", "--port", "80", "--module", module]
    if type == "agent":
        cmd.append("--remote")
    cmd_str = ",".join([f'"{c}"' for c in cmd])
    return f"""
FROM python:3.12-slim

ARG UV_VERSION="latest"
RUN apt update && apt install -y curl build-essential

# Install uv.
COPY --from=ghcr.io/astral-sh/uv:latest /uv /uvx /bin/
RUN curl -fsSL https://raw.githubusercontent.com/beamlit/toolkit/main/install.sh | BINDIR=/bin sh
WORKDIR /beamlit

# Install the application dependencies.
COPY pyproject.toml /beamlit/pyproject.toml
COPY uv.lock /beamlit/uv.lock

COPY README.m[d] /beamlit/README.md
COPY LICENS[E] /beamlit/LICENSE
COPY {settings.server.directory} /beamlit/src

RUN uv sync --no-cache --no-dev

ENV PATH="/beamlit/.venv/bin:$PATH"

ENTRYPOINT [{cmd_str}]
"""

def clean_auto_generated(
    directory: str,
    type: Literal["agent", "function"],
    deployments: list[tuple[Resource, Agent | Function]]
):
    """
    Cleans up auto-generated deployments of a specific type.

    Args:
        directory (str): Base directory containing deployments.
        type (Literal["agent", "function"]): Type of deployment to clean ("agent" or "function").
        deployments (list[tuple[Resource, Agent | Function]]): List of deployment resources and configurations.
    """

    deploy_dir = Path(directory) / f"{type}s"
    deploy_names = [d.metadata.name for (_, d) in deployments]

    if deploy_dir.exists():
        for item_dir in deploy_dir.iterdir():
            if item_dir.is_dir() and item_dir.name not in deploy_names:
                yaml_file = item_dir / f"{type}.yaml"
                if yaml_file.exists():
                    with open(yaml_file) as f:
                        try:
                            content = yaml.safe_load(f)
                            if content.get("metadata", {}).get("labels", {}).get("x-beamlit-auto-generated") == "true":
                                shutil.rmtree(item_dir)
                        except yaml.YAMLError:
                            continue

def generate_beamlit_deployment(directory: str, name: str):
    """
    Generates all necessary deployment files for Beamlit agents and functions.

    Args:
        directory (str): Target directory for generated files.
        name (str): Name identifier for the deployment.

    Creates:
        - Agent and function YAML configurations.
        - Dockerfiles for each deployment.
        - Directory structure for agents and functions.
    """
    settings = init()
    client = new_client()
    logger = getLogger(__name__)
    logger.info(f"Importing server module: {settings.server.module}")
    functions: list[tuple[Resource, Function]] = []
    agents: list[tuple[Resource, Agent]] = []
    for resource in get_resources("agent", settings.server.directory):
        agent = get_beamlit_deployment_from_resource(settings, resource)
        if name and agent.metadata.name != name:
            agent.metadata.name = slugify(name)
        if agent:
            agents.append((resource, agent))
    for resource in get_resources("function", settings.server.directory):
        function = get_beamlit_deployment_from_resource(settings, resource)
        if function:
            functions.append((resource, function))

    agents_dir = os.path.join(directory, "agents")
    functions_dir = os.path.join(directory, "functions")
    # Create directory if it doesn't exist
    os.makedirs(agents_dir, exist_ok=True)
    os.makedirs(functions_dir, exist_ok=True)
    for resource, agent in agents:
        # write deployment file
        agent_dir = os.path.join(agents_dir, agent.metadata.name)
        os.makedirs(agent_dir, exist_ok=True)
        with open(os.path.join(agent_dir, "agent.yaml"), "w") as f:
            content = get_agent_yaml(agent, functions, settings, client)
            f.write(content)
        # write dockerfile for build
        with open(os.path.join(agent_dir, "Dockerfile"), "w") as f:
            content = dockerfile("agent", resource, agent)
            f.write(content)
    for resource, function in functions:
        # write deployment file
        function_dir = os.path.join(functions_dir, function.metadata.name)
        os.makedirs(function_dir, exist_ok=True)
        with open(os.path.join(function_dir, "function.yaml"), "w") as f:
            content = get_function_yaml(function, settings, client)
            f.write(content)
        # write dockerfile for build
        with open(os.path.join(function_dir, "Dockerfile"), "w") as f:
            content = dockerfile("function", resource, function)
            f.write(content)

    clean_auto_generated(directory, "agent", agents)
    clean_auto_generated(directory, "function", functions)


"""
This module provides utility functions to format deployment configurations into YAML-compatible strings.
It includes functions to convert arguments, parameters, dictionaries, and agent chains into properly formatted JSON or YAML strings.
"""

import ast


def arg_to_list(arg: ast.List):
    value = []
    for v in arg.elts:
        value.append(format_value(v))
    return value

def format_value(v):
    """
    Formats an AST node value into its Python equivalent.

    Args:
        v (ast.AST): The AST node to format.

    Returns:
        Any: The formatted Python value.
    """
    if isinstance(v, ast.Constant):
        return v.value
    elif isinstance(v, ast.Dict):
        return arg_to_dict(v)
    elif isinstance(v, ast.List):
        return arg_to_list(v)

def arg_to_dict(arg: ast.keyword):
    """
    Converts an AST keyword argument to a dictionary.

    Args:
        arg (ast.keyword): The AST keyword argument.

    Returns:
        dict: The resulting dictionary.
    """
    value = {}
    for k, v in zip(arg.keys, arg.values):
        if isinstance(k, ast.Constant):
            value[k.value] = format_value(v)
    return value


from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="WebsocketChannel")


@_attrs_define
class WebsocketChannel:
    """WebSocket connection details

    Attributes:
        created_at (Union[Unset, str]): The date and time when the resource was created
        updated_at (Union[Unset, str]): The date and time when the resource was updated
        connection_id (Union[Unset, str]): Unique connection ID
        workspace (Union[Unset, str]): Workspace the connection belongs to
    """

    created_at: Union[Unset, str] = UNSET
    updated_at: Union[Unset, str] = UNSET
    connection_id: Union[Unset, str] = UNSET
    workspace: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        created_at = self.created_at

        updated_at = self.updated_at

        connection_id = self.connection_id

        workspace = self.workspace

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if created_at is not UNSET:
            field_dict["createdAt"] = created_at
        if updated_at is not UNSET:
            field_dict["updatedAt"] = updated_at
        if connection_id is not UNSET:
            field_dict["connection_id"] = connection_id
        if workspace is not UNSET:
            field_dict["workspace"] = workspace

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        created_at = d.pop("createdAt", UNSET)

        updated_at = d.pop("updatedAt", UNSET)

        connection_id = d.pop("connection_id", UNSET)

        workspace = d.pop("workspace", UNSET)

        websocket_channel = cls(
            created_at=created_at,
            updated_at=updated_at,
            connection_id=connection_id,
            workspace=workspace,
        )

        websocket_channel.additional_properties = d
        return websocket_channel

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import asyncio
import os
import warnings
from dataclasses import dataclass
from typing import Callable

import pydantic
import typing_extensions as t
from langchain_core.tools.base import BaseTool, ToolException

from beamlit.api.agents import list_agents
from beamlit.authentication.authentication import AuthenticatedClient
from beamlit.common.settings import get_settings
from beamlit.models import Agent, AgentChain
from beamlit.run import RunClient


class ChainTool(BaseTool):
    """
    A tool that allows chaining of agent actions. Extends LangChain's BaseTool.
    """

    client: RunClient
    handle_tool_error: bool | str | Callable[[ToolException], str] | None = True
    _cloud: bool = False
    _service_name: str | None = None

    @t.override
    def _run(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        """
        Executes the tool synchronously.

        Parameters:
            *args (Any): Positional arguments.
            **kwargs (Any): Keyword arguments.

        Returns:
            Any: The result of the tool execution.
        """
        warnings.warn(
            "Invoke this tool asynchronousely using `ainvoke`. This method exists only to satisfy standard tests.",
            stacklevel=1,
        )
        return asyncio.run(self._arun(*args, **kwargs))

    @t.override
    async def _arun(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        """
        Executes the tool asynchronously.

        Parameters:
            *args (Any): Positional arguments.
            **kwargs (Any): Keyword arguments.

        Returns:
            Any: The result of the asynchronous tool execution.
        """
        result = self.client.run(
            "agent",
            self.name,
            "POST",
            cloud=self._cloud,
            service_name=self._service_name,
            json=kwargs,
        )
        return result.text

    @t.override
    @property
    def tool_call_schema(self) -> type[pydantic.BaseModel]:
        """
        Defines the schema for tool calls based on the provided argument schema.

        Returns:
            type[pydantic.BaseModel]: The Pydantic model representing the tool call schema.
        """
        assert self.args_schema is not None  # noqa: S101
        return self.args_schema


class ChainInput(pydantic.BaseModel):
    """
    A Pydantic model representing the input structure for a chain.
    """

    inputs: str


@dataclass
class ChainToolkit:
    """
    A toolkit for managing and initializing a chain of agents.
    """

    client: AuthenticatedClient
    chain: list[AgentChain]
    _cloud: bool = False
    _service_name: str | None = None
    _chain: list[Agent] | None = None

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    def initialize(self) -> None:
        """
        Initializes the toolkit by retrieving and configuring the list of agents based on the provided chains.

        Raises:
            RuntimeError: If initialization fails due to missing agents.
        """
        """Initialize the session and retrieve tools list"""
        settings = get_settings()
        self._cloud = settings.cloud
        if self._chain is None:
            agents = list_agents.sync_detailed(
                client=self.client,
            ).parsed
            chain_enabled = [chain for chain in self.chain if chain.enabled]
            agents_chain = []
            for chain in chain_enabled:
                agent = [agent for agent in agents if agent.metadata.name == chain.name]
                agent_name = agent[0].metadata.name.upper().replace("-", "_")
                if os.getenv(f"BL_AGENT_{agent_name}_SERVICE_NAME"):
                    self._service_name = os.getenv(f"BL_AGENT_{agent_name}_SERVICE_NAME")
                if agent:
                    agent[0].spec.prompt = chain.prompt or agent[0].spec.prompt
                    agent[0].spec.description = chain.description or agent[0].spec.description
                    agents_chain.append(agent[0])
            self._chain = agents_chain

    def get_tools(self) -> list[BaseTool]:
        """
        Retrieves a list of tools corresponding to the initialized agents.

        Returns:
            list[BaseTool]: A list of initialized `ChainTool` instances.

        Raises:
            RuntimeError: If the toolkit has not been initialized.
        """
        if self._chain is None:
            raise RuntimeError("Must initialize the toolkit first")

        return [
            ChainTool(
                client=RunClient(self.client),
                name=agent.metadata.name,
                description=agent.spec.description or agent.spec.prompt or "",
                args_schema=ChainInput,
                cloud=self._cloud,
                service_name=self._service_name,
            )
            for agent in self._chain
        ]